@@ -13,21 +13,35 @@ from tqdm.auto import tqdm
 import numpy as np
 from scipy import optimize
 
-from . import gp
+from noctiluca import make_TaggedSet
+
+from .gp import GP, msd2C_fun
 from .deco import method_verbosity_patch
 
 class Fit(metaclass=ABCMeta):
     """
-    Abstract base class for MSD fits. Backbone of bayesmsd.
+    Abstract base class for MSD fits; the backbone of bayesmsd.
 
     Subclass this to implement fitting of a specific functional form of the
     MSD. See also the existing library of fits in `bayesmsd.lib`.
 
     Parameters
     ----------
-    data : `TaggedSet` of `Trajectory`
-        the data to use. Note that the current selection in the data is saved
-        internally
+    data : noctiluca.TaggedSet, pandas.DataFrame, list of numpy.ndarray
+        the data to run the fit on.
+
+        This input is handled by the `userinput.make_TaggedSet()
+        <https://noctiluca.readthedocs.io/en/latest/noctiluca.util.html#noctiluca.util.userinput.make_TaggedSet>`_
+        function of the `noctiluca` package and thus accepts a range of
+        formats. A dataset of trajectories with ``N`` loci, ``T`` frames, and
+        ``d`` spatial dimensions can be specified as list of numpy arrays with
+        dimensions ``(N, T, d)``, ``(T, d)``, ``(T,)``, where ``T`` can vary
+        between trajectories but ``N`` and ``d`` should be the same. A pandas
+        dataframe should have columns ``(particle, frame, x1, y1, z1, x2, y2,
+        z2, ...)``, where ``particle`` identifies which trajectory each entry
+        belongs to, which ``frame`` is the frame number in which it was
+        detected. For precise specs see `noctiluca.util.userinput
+        <https://noctiluca.readthedocs.io/en/latest/noctiluca.util.html#module-noctiluca.util.userinput>`_.
 
     Attributes
     ----------
@@ -41,16 +55,10 @@ class Fit(metaclass=ABCMeta):
     ss_order : {0, 1}
         steady state order. Often this will be a fixed constant for a
         particular `Fit` subclass.
-    bounds : list of (lb, ub)
-        bound for each of the parameters in the fit. This is everything the
-        class "knows" about your choice of parameters internally. The length of
-        this list is also used internally to count number of parameters, so it
-        is important that every parameter has bounds. Use ``np.inf`` and
-        ``-np.inf`` for unbounded parameters.
-    fix_values : list of (i, fix_to)
-        allows to fix some parameter values to constant or values of other
-        parameters, e.g. to allow for different behavior in different
-        dimensions, but fixing it to be equal by default. See Notes section.
+    parameters : dict
+        the parameters for this fit. Each entry should carry a sensible name
+        and be an instance of `bayesmsd.Parameter
+        <bayesmsd.parameters.Parameter>`.
     constraints : list of constraint functions
         allows to specify constraints on the parameters that will be
         implemented as smooth penalty on the likelihood. Can also take care of
@@ -72,26 +80,19 @@ class Fit(metaclass=ABCMeta):
 
     Notes
     -----
-    The `!fix_values` mechanism allows to keep some parameters fixed, or
-    express them as function of others. ``Fit.fix_values`` is a list of tuples
-    ``(i, fix_to)``, where ``i`` is the index of the parameter you want to fix,
-    c.f. `!bounds`. ``fix_to`` is either just a constant value, or a function
-    ``fix_to(params) --> float``, where ``params`` are the current parameter
-    values. Note that in this function you should not rely on any parameters
-    that are themselves to be fixed. (It would get impossible to resolve all
-    the dependencies).
-
     `!constraints` are functions with signature ``constraint(params) -->
     float``. The output is interpreted as follows:
-    - x <= 0 : infeasible; maximum penalization
-    - 0 <= x <= 1 : smooth penalization: ``penalty = exp(1/tan(pi*x))``
-    - 1 <= x : feasible; no penalization
+
+     - x <= 0 : infeasible; maximum penalization
+     - 0 <= x <= 1 : smooth penalization: ``penalty = exp(cot(pi*x))``
+     - 1 <= x : feasible; no penalization
+
     Thus, if e.g. some some function ``fun`` of the parameters should be
     constrained to be positive, you would use ``fun(params)/eps`` as the
     constraint, with ``eps`` some small value setting the tolerance region. If
     there are multiple constraints, always the strongest one is used. For
-    infeasible parameters, the likelihood function is not evaluated, but the
-    "likelihood" is just set to ``-Fit.max_penalty``.
+    infeasible parameters, the likelihood function is not evaluated; instead
+    the "likelihood" is just set to ``-Fit.max_penalty``.
 
     Note that there is a default constraint checking positivity of the
     covariance matrix. If your functional form of the MSD is guaranteed to
@@ -111,27 +112,28 @@ class Fit(metaclass=ABCMeta):
     (L-BFGS-B). The latter uses the stopping criterion ``f^k -
     f^{k+1}/max{|f^k|,|f^{k+1}|,1} <= ftol``, which is inappropriate for
     log-likelihoods (which should be optimized to fixed accuracy of O(0.1)
-    independent of the absolute value, which might be very large). We therefore
-    use Nelder-Mead by default, which does not depend on derivatives and thus
-    also has an absolute stopping criterion.
+    independent of the absolute value, which in turn might be very large). We
+    therefore use Nelder-Mead by default, which does not depend on derivatives
+    and thus also has an absolute stopping criterion.
 
-    If this function runs close to a maximum, e.g. in `Profiler` or when using
-    successive optimization steps, we can fix the problem with gradient-based
-    optimization by removing the large absolute value offset from the
-    log-likelihood. This functionality is also exposed to the end user, who can
-    overwrite the ``initial_offset`` method to give a non-zero offset together
-    with the initial values provided via `initial_params`.
+    If this function runs close to a maximum, e.g. in `Profiler
+    <bayesmsd.profiler.Profiler>` or when using successive optimization steps,
+    we can fix the problem with gradient-based optimization by removing the
+    large absolute value offset from the log-likelihood. This functionality is
+    also exposed to the end user, who can overwrite the ``initial_offset``
+    method to give a non-zero offset together with the initial values provided
+    via `initial_params`.
     """
     def __init__(self, data):
-        self.data = data
-        self.data_selection = data.saveSelection()
-        self.d = data.map_unique(lambda traj: traj.d)
+        self.data = make_TaggedSet(data)
+        self.data_selection = self.data.saveSelection()
+        self.d = self.data.map_unique(lambda traj: traj.d)
         self.T = max(map(len, self.data))
 
         # Fit properties
-        self.ss_order = 0
-        self.bounds = [] # List of (lb, ub), to be passed to optimize.minimize. len(bounds) is used to count parameters!
-        self.fix_values = [] # will be appended to the list passed to run()
+        self.ss_order = None
+        self.parameters = {} # should be dict of parameters.Parameter instances
+        # `params` arguments will then be dicts as well
 
         # Each constraint should be a callable constr(params) -> x. We will apply:
         #   x <= 0                : infeasible. Maximum penalization
@@ -157,22 +159,24 @@ class Fit(metaclass=ABCMeta):
         Definition of MSD in terms of parameters
 
         This is the core of the fit definition. It should give a list of tuples
-        as required by `gp.ds_logL`
+        as required by `GP.ds_logL <bayesmsd.gp.GP.ds_logL>`
 
         Parameters
         ----------
-        params : np.ndarray
-            the current parameters
+        params : dict
+            the current parameter values
 
         Returns
         -------
         list of tuples (msd, m)
-            for the definition of `!msd`, use of the `MSDfun` decorator is
-            recommended.
+            for the definition of `!msd`, use of the `MSDfun
+            <bayesmsd.deco.MSDfun>` decorator (and potentially also `imaging
+            <bayesmsd.deco.imaging>`) is recommended.
 
         See also
         --------
-        gp.ds_logL, MSDfun, imaging
+        GP.ds_logL <bayesmsd.gp.GP.ds_logL>, MSDfun <bayesmsd.deco.MSDfun>,
+        imaging <bayesmsd.deco.imaging>
         """
         raise NotImplementedError # pragma: no cover
 
@@ -186,8 +190,8 @@ class Fit(metaclass=ABCMeta):
 
         Parameters
         ----------
-        params : np.ndarray
-            the current parameters
+        params : dict
+            the current parameter values
 
         Returns
         -------
@@ -206,9 +210,9 @@ class Fit(metaclass=ABCMeta):
 
         Returns
         -------
-        params : np.ndarray
-            initial values for the parameters. Should satisfy ``len(params) ==
-            len(self.bounds)``.
+        params : dict
+            initial values for the parameters. Should have the same keys as
+            ``self.parameters``.
         """
         raise NotImplementedError # pragma: no cover
 
@@ -257,7 +261,7 @@ class Fit(metaclass=ABCMeta):
         done = []
         for msd, _ in self.params2msdm(params):
             if msd not in done:
-                min_ev = np.min(np.linalg.eigvalsh(gp.msd2C_fun(msd, np.arange(self.T), self.ss_order)))
+                min_ev = np.min(np.linalg.eigvalsh(msd2C_fun(msd, np.arange(self.T), self.ss_order)))
                 scores.append(min_ev / min_ev_okay)
                 done.append(msd)
 
@@ -265,16 +269,15 @@ class Fit(metaclass=ABCMeta):
     
     ### General machinery, usually won't need overwriting ###
 
-    def MSD(self, fitres, dt=None):
+    def MSD(self, params, dt=None):
         """
         Return the MSD evaluated at dt. Convenience function
 
         Parameters
         ----------
-        fitres : dict
-            the output of a previous fit run. Specifically, this should be a
-            dict having an entry ``'params'`` that is a 1d ``np.array``
-            containing the parameter values.
+        params : dict
+            the current parameter values. Like the ``'params'`` entry of the
+            output dict from any fit.
         dt : array-like
             the time lags (in frames) at which to evaluate the MSD. If left
             unspecified, we return a callable MSD function
@@ -286,18 +289,18 @@ class Fit(metaclass=ABCMeta):
         callable or np.array
             the MSD function (summed over all dimensions), evaluated at `!dt` if provided.
         """
-        def msdfun(dt, params=fitres['params'], **kwargs):
+        def msdfun(dt, params=params, **kwargs):
             msdm = self.params2msdm(params)
             return np.sum([msd(dt, **kwargs) for msd, m in msdm], axis=0)
 
         if dt is None:
             # Write proper signature to docstring
-            single_msdfun = self.params2msdm(fitres['params'])[0][0]
+            single_msdfun = self.params2msdm(params)[0][0]
             if hasattr(single_msdfun, '_kwargstring'):
                 msdfun.__doc__ = f"""full signature:
 
 msdfun(dt,
-       params={fitres['params']},
+       params={params},
        {single_msdfun._kwargstring},
       )
 
@@ -313,17 +316,29 @@ msdfun(dt,
 
         Parameters
         ----------
-        params : np.ndarray
+        params : dict
 
         Returns
         -------
         float
+            anything < 0 indicates infeasibility, so don't even evaluate the
+            likelihood
         """
         x = np.inf
+
+        # Check parameter bounds
+        # We can just detect infeasibility; if a soft bound is needed, implement
+        # a constraint
+        for name, P in self.parameters.items():
+            if (   P.bounds[0] > params[name]
+                or P.bounds[1] < params[name]):
+                return -1 # infeasible
+
+        # Check inequality constraints
         for constraint in self.constraints:
             x = min(constraint(params), x)
             if x <= 0:
-                return -1 # unfeasible
+                return -1 # infeasible
 
         if x >= 1:
             return 0
@@ -336,165 +351,237 @@ msdfun(dt,
 
     def expand_fix_values(self, fix_values=None):
         """
-        Preprocessing for fixed parameters. Mostly internal use.
+        Assemble full list of values to fix (internal + given)
 
-        This function concatenates the internal ``self.fix_values`` and the
-        given ``fix_values``, and makes sure that they all conform to the
-        format ``(i, function)``, i.e. it converts fixed values given as
-        constants to (trivial) functions.
+        Merge the given `!fix_values` with the parameter-level specifications
+        and ensure that all entries are appropriate callables (can also be
+        specified as numerical constants or other parameter names)
 
         Parameters
         ----------
-        fix_values : list of tuples (i, fix_to), optional
-            values to fix, beyond what's already in ``self.fix_values``
+        fix_values : dict, optional
+            values to fix, beyond what's already in
+            ``self.parameters[...].fix_to``
 
         Returns
         -------
-        fix_values : list
-            same as input, plus internal ``self.fix_values`` and constants
-            resolved
+        fix_values : dict
+            same as input, plus internals, and constants resolved
 
-        See also
-        --------
-        Fit.fix_values, get_value_fixer
+        Notes
+        -----
+        Parameter resolution order:
+
+         + free parameters
+         + fixed to constant
+         + fixed to other parameter by name
+         + fixed to callable
+
+        So you cannot fix to another parameter that is itself fixed to a
+        callable. You can use ``parameter.fix_to = other_parameter.fix_to``
+        instead to just copy the function over instead of fixing by name. This
+        helps prevent undetectable cycles in the fixes.
         """
-        # Converts fix_values to useable format (i, function)
-        # Also appends self.fix_values
-        if fix_values is None:
-            fix_values = []
-        fix_values = fix_values + self.fix_values
-        
-        ifix = [i for i, _ in fix_values]
-        _, ind = np.unique(ifix, return_index=True)
-        full = []
-        for i in ind:
-            ip, fun = fix_values[i]
-            if not callable(fun):
-                fun = lambda x, val=fun : val # `fun` is (initially) of course a misnomer in this case
+        # Merge input and fit-internal fix_values (input takes precedence)
+        # Note: using the dict here ensures uniqueness (only one fix per
+        # parameter)
+        fix_values_in = fix_values if fix_values is not None else {}
+        fix_values = {name : param.fix_to
+                      for name, param in self.parameters.items()
+                      if param.fix_to is not None}
+        fix_values.update(fix_values_in)
 
-            full.append((ip, fun))
-        return full
+        # Keep track of which parameter is resolved how, such that we can give
+        # a good resolution order later
+        unfixed = [name for name in self.parameters if name not in fix_values]
+        to_constant = [] # just names
+        to_other = []    # (name, resolves_to)
+        to_callable = [] # just names
 
-    def number_of_fit_parameters(self, fix_values=None):
+        # Convert constants (name or numerical) to callables
+        for name, fix_to in fix_values.items():
+            if callable(fix_to):
+                to_callable.append(name)
+            else:
+                if fix_to in self.parameters:
+                    to_other.append((name, fix_to))
+                    fix_values[name] = lambda x, name=fix_to : x[name]
+                else:
+                    to_constant.append(name)
+                    fix_values[name] = lambda x, val=fix_to : val
+
+        # Determine resolution order:
+        # - constants go first, in undetermined order
+        # - then resolve identifications recursively
+        # - finally the general callables
+        resolution_order = unfixed + to_constant
+        while len(to_other) > 0:
+            cache_len = len(to_other)
+            i = 0
+            while i < len(to_other):
+                if to_other[i][1] in resolution_order:
+                    resolution_order.append(to_other.pop(i)[0])
+                else:
+                    i += 1
+
+            if len(to_other) == cache_len:
+                # Could not insert any more of the identifications, so there
+                # must be a cycle in the remaining fixes (or something depends
+                # on a parameter that's fixed by callable)
+                self.vprint(1, "Left-overs from resolution order determination:\n"
+                              f"resolution_order = {resolution_order}\n"
+                              f"        to_other = {to_other}\n"
+                              f"     to_callable = {to_callable}\n")
+                raise RuntimeError("Could not determine resolution order of parameter fixes.")
+
+        resolution_order += to_callable
+
+        # Remove unfixed parameters
+        assert len(resolution_order) == len(self.parameters)
+        assert resolution_order[:len(unfixed)] == unfixed
+        resolution_order = resolution_order[len(unfixed):]
+
+        return fix_values, resolution_order, unfixed
+
+    def independent_parameters(self, fix_values=None):
         """
-        Get the number of independent fit parameters
-
-        The number of independent fit parameters is the full number of
-        parameters (``len(params)``) minus the number of parameters that are
-        fixed via the ``fix_values`` mechanism. This function is provided for
-        convenience.
+        Return the names of the independent (not fixed) parameters
 
         Parameters
         ----------
-        fix_values : list of tuples (i, fix_to), optional
+        fix_values : dict
             should be the same as was / will be handed to `run`
 
         Returns
         -------
-        int
-            the number of fit parameters
+        list of str
+            a list with the names of the independent parameters, i.e. those
+            that are not fixed to some other value
         """
-        n_fixed = len(self.expand_fix_values(fix_values)) # also ensures uniqueness of indices
-        return len(self.bounds) - n_fixed
+        if fix_values is None:
+            fix_values = {}
+
+        return [name for name in self.parameters if (name not in fix_values and self.parameters[name].fix_to is None)]
                 
-    def get_value_fixer(self, fix_values=None):
+    class MinTarget:
         """
-        Create a function that resolves fixed parameters
+        Helper class; acts as cost function for optimization.
 
-        Parameters
-        ----------
-        fix_values : list of tuples (i, fix_to), optional
-            values to fix, beyond what's already in ``self.fix_values``
-
-        Returns
-        -------
-        fixer : callable
-            a function with signature ``fixer(params) --> params``, where in
-            the output array all parameters fixes are applied.
-        """
-        fix_values = self.expand_fix_values(fix_values)
-
-        n_params = len(self.bounds)
-        is_fixed = np.zeros(n_params, dtype=bool)
-        for i, _ in fix_values:
-            is_fixed[i] = True
-        
-        def value_fixer(params, fix_values=fix_values, is_fixed=is_fixed, n_params=n_params):
-            if len(params) == n_params:
-                fixed_params = params.copy()
-            elif len(params) + np.sum(is_fixed) == n_params:
-                fixed_params = np.empty(n_params, dtype=float)
-                fixed_params[:] = np.nan
-                fixed_params[~is_fixed] = params
-            else: # pragma: no cover
-                raise RuntimeError(f"Did not understand how to fix up parameters: len(params) = {len(params)}, sum(is_fixed) = {np.sum(is_fixed)}, n_params = {n_params}")
-
-            for ip, fixfun in fix_values:
-                fixed_params[ip] = fixfun(fixed_params)
-            
-            return fixed_params
-
-        return value_fixer
-    
-    def get_min_target(self, offset=0, fix_values=None, do_fixing=True):
-        """
-        Define the minimization target (negative log-likelihood)
+        Beyond acting as the actual cost function through implementation of the
+        ``()`` operator, this class also defines the conversion between
+        parameter dicts (where parameters are named and unordered) and
+        parameter arrays (which are used for optimization). Most of this is for
+        internal use.
 
         Parameters
         ----------
+        fit : Fit
+            the `Fit` object that this target is associated with.
+        fix_values : dict
+            any additional fixes for parameters, beyond what's already in
+            ``fit.parameters[...].fix_to``.
         offset : float
-            constant to subtract from log-likelihood. See Notes section of
-            class doc.
-        fix_values : list of tuples (i, fix_to)
-            values to fix, beyond what's already in ``self.fix_values``
-        do_fixing : bool
-            set to ``False`` to prevent the minimization target from resolving
-            any of the parameter fixes (by default). Might be useful when
-            exploring parameter space.
+            the global offset to subtract; see Notes section of `Fit`.
 
-        Returns
-        -------
-        min_target : callable
-            function with signature ``min_target(params) --> float``.
-
-        Notes
-        -----
-        The returned ``min_target`` takes additional keyword arguments:
-
-        - ``just_return_full_params`` : bool, ``False`` by default. If
-          ``True``, don't calculate the actual target function, just return the
-          parameter values after fixing
-        - ``do_fixing`` : bool, defaults to `!do_fixing` as given to
-          `get_min_target`.
-        - fixer, offset : just handed over as arguments for style (scoping)
-
-        See also
-        --------
-        run
+        Attributes
+        ----------
+        fit : Fit
+            the `Fit` object that this target is associated with.
+        fix_values : dict
+            the complete set of values to keep fixed; keys are parameter names,
+            values are (guaranteed to be) callables.
+        fix_values_resolution_order : list of str
+            list of parameter names. The fixes in `!fix_values` will be
+            executed in this order.
+        param_names : list of str
+            list of parameter names that remain independent; this determines
+            the order of entries when converting parameter dicts to arrays.
+        offset : float
+            constant to be subtracted from the target (i.e. added to the
+            likelihood). Can be used to ensure that the minimum value of the
+            target is close to 0.
         """
-        fixer = self.get_value_fixer(fix_values)
-        
-        def min_target(params, just_return_full_params=False,
-                       do_fixing=do_fixing, fixer=fixer, offset=offset,
-                       ):
-            if do_fixing:
-                params = fixer(params)
-            if just_return_full_params:
-                return params
+        def __init__(self, fit, fix_values=None, offset=0):
+            self.fit = fit
 
-            penalty = self._penalty(params)
+            # See class docstring
+            fv = self.fit.expand_fix_values(fix_values)
+            self.fix_values                  = fv[0]
+            self.fix_values_resolution_order = fv[1]
+            self.param_names                 = fv[2]
+
+            self.offset = offset
+
+        def params_array2dict(self, params_array):
+            """
+            Convert a parameter array to dict
+
+            Parameters
+            ----------
+            params_array : np.ndarray
+
+            Returns
+            -------
+            dict
+
+            See also
+            --------
+            params_dict2array
+            """
+            params = dict(zip(self.param_names, params_array))
+            for name in self.fix_values_resolution_order:
+                params[name] = self.fix_values[name](params)
+
+            return params
+
+        def params_dict2array(self, params):
+            """
+            Convert a parameter dict to array
+
+            Parameters
+            ----------
+            params_dict : dict
+
+            Returns
+            -------
+            np.ndarray
+
+            See also
+            --------
+            params_array2dict
+            """
+            return np.array([params[name] for name in self.param_names])
+
+        def __call__(self, params_array):
+            """
+            Actual minimzation target
+
+            This function evaluates penalty, prior, and likelihood for the
+            current parameters and returns the negative of their sum. If
+            ``penalty < 0``, instead of evaluating the likelihood, we just
+            return the maximum penalization.
+
+            Parameters
+            ----------
+            params_array : np.ndarray
+
+            Returns
+            -------
+            float
+            """
+            params = self.params_array2dict(params_array)
+
+            penalty = self.fit._penalty(params)
             if penalty < 0: # infeasible
-                return self.max_penalty
+                return self.fit.max_penalty
             else:
-                return -gp.ds_logL(self.data,
-                                   self.ss_order,
-                                   self.params2msdm(params),
+                return -GP.ds_logL(self.fit.data,
+                                   self.fit.ss_order,
+                                   self.fit.params2msdm(params),
                                   ) \
-                       - self.logprior(params) \
+                       - self.fit.logprior(params) \
                        + penalty \
-                       - offset
-
-        return min_target
+                       - self.offset
 
     @method_verbosity_patch
     def run(self,
@@ -512,9 +599,8 @@ msdfun(dt,
         ----------
         init_from : dict
             initial point for the fit, as a dict with fields ``'params'`` and
-            ``'logL'``, like the ones this function returns. If you just want
-            to initialize from a certain parameter point, you can set
-            ``init_from['logL'] = 0``.
+            ``'logL'``, like the ones this function returns. The ``'logL'``
+            entry is optional; you can also leave it out or set to 0.
         optimization_steps : tuple
             successive optimization steps to perform. Entries should be
             ``'simplex'`` for Nelder-Mead, ``'gradient'`` for gradient descent,
@@ -523,14 +609,14 @@ msdfun(dt,
         maxfev : int or None
             limit on function evaluations for ``'simplex'`` or ``'gradient'``
             optimization steps
-        fix_values : list of tuples (i, fix_to)
+        fix_values : dict
             can be used to keep some parameter values fixed or express them as
             function of the other parameters. See class doc for more details.
         full_output : bool
             Set to ``True`` to return the output dict (c.f. Returns) and the
             full output from ``scipy.optimize.minimize`` for each optimization
             step. Otherwise (``full_output == False``, the default) only the
-            output dict from the ultimate run is returned.
+            output dict from the final run is returned.
         show_progress : bool
             display a `!tqdm` progress bar while fitting
         verbosity : {None, 0, 1, 2, 3}
@@ -551,19 +637,21 @@ msdfun(dt,
         
         # Initial values
         if init_from is None:
-            p0 = self.initial_params()
-            total_offset = self.initial_offset()
+            initial_params = self.initial_params()
+            initial_offset = self.initial_offset()
         else:
-            p0 = deepcopy(init_from['params'])
-            total_offset = -init_from['logL']
-        
-        # Adjust for fixed values
-        ifix = [i for i, _ in self.expand_fix_values(fix_values)]
-        bounds = deepcopy(self.bounds)
-        for i in sorted(ifix, reverse=True):
-            del bounds[i]
-            p0[i] = np.nan
-        p0 = p0[~np.isnan(p0)]
+            initial_params = deepcopy(init_from['params'])
+            try:
+                initial_offset = -init_from['logL']
+            except KeyError:
+                initial_offset = 0
+
+        # Set up the minimization target
+        # also allows us to convert initial_params to appropriate array
+        min_target = Fit.MinTarget(self, fix_values, initial_offset)
+
+        p0 = min_target.params_dict2array(initial_params)
+        bounds = [self.parameters[name].bounds for name in min_target.param_names]
         
         # Set up progress bar
         bar = tqdm(disable = not show_progress)
@@ -608,12 +696,11 @@ msdfun(dt,
                                  )
                     kwargs.update(step)
                     
-                min_target = self.get_min_target(offset=total_offset, fix_values=fix_values)
                 try:
                     fitres = optimize.minimize(min_target, p0, **kwargs)
-                except gp.BadCovarianceError as err: # pragma: no cover
+                except GP.BadCovarianceError as err: # pragma: no cover
                     self.vprint(2, "BadCovarianceError:", err)
-                    fitres = lambda: None
+                    fitres = lambda: None # hack: lambdas allow free assignment of attributes
                     fitres.success = False
                 
                 if not fitres.success:
@@ -621,11 +708,11 @@ msdfun(dt,
                     self.vprint(1, '\n', fitres)
                     raise RuntimeError("Fit failed at step {:d}: {:s}".format(istep, step))
                 else:
-                    all_res.append(({'params' : min_target(fitres.x, just_return_full_params=True),
-                                     'logL' : -(fitres.fun+total_offset),
+                    all_res.append(({'params' : min_target.params_array2dict(fitres.x),
+                                     'logL' : -(fitres.fun+min_target.offset),
                                     }, fitres))
                     p0 = fitres.x
-                    total_offset += fitres.fun
+                    min_target.offset += fitres.fun
                     
         bar.close()
         
