@@ -1,11 +1,12 @@
 """
 Implementation of the `Profiler` class
 
-This is used to "explore the posterior" after running a `Fit <bayesmsd.Fit>`,
-specifically to find the credible intervals.
+This is used to "explore the posterior" after running a `Fit
+<bayesmsd.fit.Fit>`, specifically to find the credible intervals.
 
 Example
 -------
+
 >>> fit = <some bayesmsd.Fit subclass, c.f. bayesmsd.lib>
 ... profiler = bayesmsd.Profiler(fit)
 ... mci = profiler.find_MCI() # mci = max posterior estimate & credible
@@ -15,6 +16,9 @@ See also
 --------
 bayesmsd, Profiler, Fit <bayesmsd.fit.Fit>
 """
+import functools
+from copy import deepcopy
+
 from tqdm.auto import tqdm
 
 import numpy as np
@@ -24,9 +28,9 @@ class Profiler():
     """
     Exploration of the posterior after finding the point estimate
 
-    This class provides a top layer on top of `Fit`, enabling more
-    comprehensive exploration of the posterior after finding the (MAP) point
-    estimate. Generally, it operates in two modes:
+    This class provides a layer on top of `Fit`, enabling more comprehensive
+    exploration of the posterior after finding the (MAP) point estimate.
+    Generally, it operates in two modes:
 
     - conditional posterior: wiggle each individual parameter, keeping all
       others fixed to the point estimate values, thus calculating conditional
@@ -39,12 +43,11 @@ class Profiler():
 
     From a Bayesian point of view, beyond conditional and profile posterior,
     the actually interesting quantity is of course the marginal posterior. This
-    is best obtained by sampling the posterior by MCMC (see module description
-    for an example implementation of such a sampler). Still, conditional or
-    profile posterior often can give a useful overview over the shape of the
-    posterior. The conditional posterior is also great for setting MCMC step
-    sizes. Profile posteriors are of course significantly more expensive than
-    conditionals.
+    is best obtained by sampling the posterior by MCMC (see Examples &
+    Tutorials). Still, conditional or profile posterior often give a useful
+    overview over the shape of the posterior. The conditional posterior is also
+    great for setting MCMC step sizes. Profile posteriors are of course
+    significantly more expensive than conditionals.
 
     At the end of the day, this class moves along either profile or conditional
     posterior until it drops below a given cutoff, and then gives the lower
@@ -53,17 +56,18 @@ class Profiler():
     "confidence level" 1-α.
 
     We follow a two-step procedure to find the lower and upper bounds: first,
-    we move out from the point estimate with a fixed step size (additively or
-    multiplicatively) until the posterior drops below the cutoff. Within the
-    thus defined bracket, we then find the actual bound to the desired accuracy
-    by bisection. If the first step fails (i.e. the posterior does not drop
-    below the cutoff far from the point estimate) the corresponding parameter
-    direction is unidentifiable and the bound will be set to infinity
+    we move out from the point estimate with a fixed step size (in the space
+    defined by `!Parameter.linearization`) until the posterior drops below the
+    cutoff. Within the thus defined bracket, we then find the actual bound to
+    the desired accuracy by bisection. If the first step fails (i.e. the
+    posterior does not drop below the cutoff far from the point estimate) the
+    corresponding parameter direction is unidentifiable and the credible
+    interval edge will be set to infinity.
     
     Note that this class will also take care of the initial point estimate, if
     necessary.
 
-    The main point of entry for the user is the `find_MCI` method, that
+    The main point of entry for the user is the `find_MCI` method, which
     performs the sweeps described above. Further, `run_fit` might be useful if
     you just need the point estimate.
 
@@ -77,31 +81,6 @@ class Profiler():
         the "confidence level" to use for the likelihood ratio cutoff
     conf_tol : float
         acceptable tolerance in the confidence level
-    bracket_strategy : 'auto', dict, or list of dict
-        specifies the strategy to use during the first step (pushing the
-        parameter out to find a proper bracket). For full control, specify a
-        list of dicts (one for each parameter) with the structure
-        ``dict(multiplicative=(bool), step=(float),
-        nonidentifiable_cutoffs=(low, high))``, where ``multiplicative``
-        indicates whether propagation is ``param <-- param*step`` or ``param
-        <-- param + step``. Correspondingly, ``step`` should be ``>1`` for
-        ``multiplicative == True``, and ``step > 0`` for ``multiplicative ==
-        False``. Finally, ``nonidentifiability_cutoffs`` specifies how far to
-        search before the direction is considered unidentifiable. This is
-        measured in "total stepsize" taken in either direction, so e.g. for a
-        multiplicative strategy the default is ``nonidentifiability_cutoffs =
-        (10, 10)``, meaning we search by a factor of 10 around the point
-        estimate in either direction.
-        
-        Instead of specifying a list of dicts, you can also just give a single
-        dict that will be applied to all parameters. Finally, by default this
-        argument is set to ``'auto'``, which means that the strategy is
-        determined from the bounds in `!fit`. If the lower bound is positive,
-        the strategy is multiplicative, else additive.
-    bracket_step : float
-        global default for the ``'step'`` parameter if ``bracket_strategy ==
-        'auto'``. Additive strategies use ``step = point_estimate *
-        (bracket_step - 1)``.
     max_fit_runs : int
         an upper bound for how often to re-run the fit (when profiling).
     max_restarts : int
@@ -119,9 +98,9 @@ class Profiler():
     ----------
     fit : Fit
         see Parameters
-    min_target_from_fit : callable
+    min_target_from_fit : Fit.MinTarget
         the minimization target of ``self.fit``
-    ress : list
+    ress : dict
         storage of all evaluated parameter points. Each entry corresponds to a
         parameter dimension and contains a list of dicts that were obtained
         while sweeping that parameter. The individual entries are dicts like
@@ -134,13 +113,10 @@ class Profiler():
         the target decline in the posterior value we're searching for
     LR_interval : [float, float]
         acceptable interval corresponding to `!conf_tol`
-    iparam : int
-        the index of the parameter currently being sweeped. This is a class
-        attribute mainly for readability of the code.
+    cur_param : str
+        the name of the parameter currently being sweeped.
     profiling : bool
         see Parameters
-    bracket_strategy, bracket_step : list, float
-        see Parameters
     max_fit_runs : int
         see Parameters
     run_count : int
@@ -151,7 +127,8 @@ class Profiler():
         see Parameters
     restart_on_better_point_estimate : bool
         whether to restart upon finding a better point estimate. Might make
-        sense to disable if your posterior is rugged
+        sense to disable if your posterior is rugged, to avoid restarting all
+        the time.
     bar : tqdm progress bar
         the progress bar showing successive fit evaluations
 
@@ -162,12 +139,6 @@ class Profiler():
     def __init__(self, fit,
                  profiling=True,
                  conf=0.95, conf_tol=0.001,
-                 bracket_strategy='auto', # 'auto', dict(multiplicative=(bool),
-                                          #              step=(float>1),
-                                          #              nonidentifiable_cutoffs=[(low, high)],
-                                          #             ),
-                                          #         or list of such (one for each parameter)
-                 bracket_step=1.2, # global default, but would be overridden by specifying bracket_strategy
                  max_fit_runs=100,
                  max_restarts=10,
                  verbosity=1, # 0: print nothing, 1: print warnings, 2: print everything, 3: debugging
@@ -178,23 +149,20 @@ class Profiler():
         self.verbosity = verbosity
 
         self.fit = fit
-        self.min_target_from_fit = fit.get_min_target()
+        self.min_target_from_fit = self.fit.MinTarget(self.fit)
 
-        self.ress = [[] for _ in range(len(self.fit.bounds))] # one for each iparam
+        self.ress = {name : [] for name in self.fit.independent_parameters()}
         self.point_estimate = None
         
         self.conf = conf
         self.conf_tol = conf_tol
 
-        self.iparam = None
-        if fit.number_of_fit_parameters() > 1:
-            self.profiling = profiling # also sets self.LR_interval and self.LR_target
-        else:
-            self.vprint(1, "Cannot profile with a single fit parameter; setting profiling = False")
+        self.cur_param = None
+        if profiling and len(self.fit.independent_parameters()) == 1:
+            self.vprint(1, "Cannot profile with a single independent fit parameter; setting profiling = False")
             self.profiling = False
-
-        self._bracket_strategy_input = bracket_strategy # see expand_bracket_strategy()
-        self.bracket_step = bracket_step
+        else:
+            self.profiling = profiling # also sets self.LR_interval and self.LR_target
 
         self.restart_on_better_point_estimate = True
         
@@ -221,60 +189,12 @@ class Profiler():
         if self.profiling:
             dof = 1
         else:
-            n_params = len(self.fit.bounds)
-            n_fixed = len(self.fit.fix_values)
-            dof = n_params - n_fixed
+            dof = len(self.fit.independent_parameters())
             
         self.LR_interval = [stats.chi2(dof).ppf(self.conf-self.conf_tol)/2,
                             stats.chi2(dof).ppf(self.conf+self.conf_tol)/2]
         self.LR_target = np.mean(self.LR_interval)
         
-    
-    def expand_bracket_strategy(self):
-        """
-        Preprocessor for the `!bracket_strategy` attribute
-        """
-        # We need a point estimate to set up the additive scheme, so this can't be in __init__()
-        if self.point_estimate is None: # pragma: no cover
-            raise RuntimeError("Cannot set up brackets without point estimate")
-            
-        if self._bracket_strategy_input == 'auto':
-            assert self.bracket_step > 1
-            
-            self.bracket_strategy = []
-            for iparam, bounds in enumerate(self.fit.bounds):
-                multiplicative = bounds[0] > 0
-                if multiplicative:
-                    step = self.bracket_step
-                    nonidentifiable_cutoffs = [10, 10]
-                else:
-                    pe = self.point_estimate['params'][iparam]
-                    diff_bounds = np.diff(bounds)[0]
-
-                    if np.isfinite(diff_bounds):
-                        step = diff_bounds / 20 * self.bracket_step
-                        nonidentifiable_cutoffs = [30, 30] # irrelevant, bounds will kick in before
-
-                    elif np.abs(pe) < 1e-10: # point estimate == 0, so no reference scale
-                        step = 1
-                        nonidentifiable_cutoffs = [10, 10]
-
-                    else:
-                        step = np.abs(pe)*(self.bracket_step - 1)
-                        nonidentifiable_cutoffs = 2*[2*np.abs(pe)]
-                
-                self.bracket_strategy.append({
-                    'multiplicative'          : multiplicative,
-                    'step'                    : step,
-                    'nonidentifiable_cutoffs' : nonidentifiable_cutoffs,
-                })
-        elif isinstance(self._bracket_strategy_input, dict):
-            assert self._bracket_strategy_input['step'] > (1 if self._bracket_strategy_input['multiplicative'] else 0)
-            self.bracket_strategy = len(self.fit.bounds)*[self._bracket_strategy_input]
-        else:
-            assert isinstance(self._bracket_strategy_input, list)
-            self.bracket_strategy = self._bracket_strategy_input
-        
     class FoundBetterPointEstimate(Exception):
         pass
     
@@ -286,6 +206,7 @@ class Profiler():
         restarts. So a function decorated with this decorator can just raise
         this exception and will then be restarted properly.
         """
+        @functools.wraps(fun)
         def decorated_fun(self, *args, **kwargs):
             for restarts in range(self.max_restarts_per_parameter):
                 try:
@@ -303,7 +224,7 @@ class Profiler():
                     # sense to keep the old results, since the parameters are different
                     if not self.profiling:
                         fit_kw['init_from'] = self.best_estimate
-                        self.ress = [[] for _ in range(len(self.fit.bounds))]
+                        self.ress = {name : [] for name in self.fit.independent_parameters()}
                         self.point_estimate = None
 
                     # Get a new point estimate, starting from the better one we found
@@ -352,7 +273,7 @@ class Profiler():
             return None
         else:
             best = self.point_estimate
-            for ress in self.ress:
+            for _, ress in self.ress.items():
                 try:
                     candidate = ress[np.argmax([res['logL'] for res in ress])]
                 except ValueError: # argmax([])
@@ -443,14 +364,14 @@ class Profiler():
             # check. It ended up being problematic: when we are profiling, we
             # initialize to the closest available previous point, which of
             # course will not fulfill the constraint on the parameter of
-            # interest. It's associated likelihood (which we would compare
+            # interest. Its associated likelihood (which we would compare
             # against) thus can be better than any possible value that fulfills
             # the constraint. Long story short: no sanity check here.
         
         if is_new_point_estimate:
             self.point_estimate = res
         else:
-            self.ress[self.iparam].append(res)
+            self.ress[self.cur_param].append(res)
             self.check_point_estimate_against(res)
 
         if self.bar is not None:
@@ -479,18 +400,17 @@ class Profiler():
         --------
         profile_likelihood
         """
-        ress = self.ress[self.iparam] + [self.point_estimate]
+        ress = self.ress[self.cur_param] + [self.point_estimate]
         
-        values = np.array([res['params'][self.iparam] for res in ress])
+        values = np.array([res['params'][self.cur_param] for res in ress])
         if val in values:
-            i = np.argmax([res['logL'] for res in ress if res['params'][self.iparam] == val])
+            i = np.argmax([res['logL'] for res in ress if res['params'][self.cur_param] == val])
             return ress[i]
         
-        if self.bracket_strategy[self.iparam]['multiplicative']:
-            distances = np.abs([np.log(val/res['params'][self.iparam]) for res in ress])
-        else:
-            distances = np.abs([val-res['params'][self.iparam] for res in ress])
-            
+        # Calculate distances
+        # Note that values[-1] is the point estimate
+        distances = self.fit.parameters[self.cur_param].linearization.distance(values[-1], val, values)
+
         if direction is not None:
             distances[np.sign(values - val) != direction] = np.inf
             if not np.any(np.isfinite(distances)):
@@ -510,7 +430,7 @@ class Profiler():
         Parameters
         ----------
         value : float
-            value of the current parameter of interest (``self.iparam``)
+            value of the current parameter of interest (``self.cur_param``)
         init_from : dict or 'closest'
             from where to start the optimization (only relevant if
             ``self.profiling``). Set to 'closest' to use
@@ -526,25 +446,24 @@ class Profiler():
                 init_from = self.find_closest_res(value)
 
             self.run_fit(init_from = init_from,
-                         fix_values = [(self.iparam, value)],
+                         fix_values = {self.cur_param : value},
                          is_new_point_estimate = False,
                         )
         else:
-            new_params = self.point_estimate['params'].copy()
-            new_params[self.iparam] = value
-            minus_logL = self.min_target_from_fit(new_params)
-                
+            new_params = deepcopy(self.point_estimate['params'])
+            new_params[self.cur_param] = value
+            new_params_array = self.min_target_from_fit.params_dict2array(new_params)
+            minus_logL = self.min_target_from_fit(new_params_array)
+
             if self.bar is not None:
                 self.bar.update() # pragma: no cover
             
-            self.ress[self.iparam].append({'logL' : -minus_logL, 'params' : new_params})
-            self.check_point_estimate_against(self.ress[self.iparam][-1])
+            self.ress[self.cur_param].append({'logL' : -minus_logL, 'params' : new_params})
+            self.check_point_estimate_against(self.ress[self.cur_param][-1])
             
-        return self.ress[self.iparam][-1]['logL']
+        return self.ress[self.cur_param][-1]['logL']
     
-    def iterate_bracket_point(self, x0, pL, direction,
-                              step = None,
-                             ):
+    def find_bracket_point(self, direction):
         """
         First step of the procedure ("bracketing")
 
@@ -555,14 +474,8 @@ class Profiler():
 
         Parameters
         ----------
-        x0, pL : float
-            the parameter value and associated posterior value to start from.
-            Should be from the point estimate
         direction : {-1, 1}
             which direction to go in
-        step : float or None
-            use to override ``self.bracket_strategy[self.iparam]['step']``,
-            which is used by default.
 
         Returns
         -------
@@ -575,48 +488,46 @@ class Profiler():
         --------
         initial_bracket_points
         """
-        self.expand_bracket_strategy()
-        if step is None:
-            step = self.bracket_strategy[self.iparam]['step']
-            
-        bracket_param = 1 if self.bracket_strategy[self.iparam]['multiplicative'] else 0
-        p = x0
+        # Input processing
+        assert self.point_estimate is not None
+        assert direction in {-1, 1}
+        idir = int((direction+1)/2)
+        param = self.fit.parameters[self.cur_param]
+
+        # Set up
+        pe = self.point_estimate['params'][self.cur_param]
         pL_thres = self.point_estimate['logL'] - self.LR_target
-        hit_bound = False
-        while pL > pL_thres and not hit_bound:
-            self.vprint(3, "bracketing: {:.3f} > {:.3f} @ {}".format(pL, pL_thres, p))
+
+        # Run
+        for n in range(1, param.max_linearization_moves[idir]+1):
+            # Next move
+            p = param.linearization.move(pe, pe, direction*n)
             
-            # Check identifiability cutoff
-            ib = int((1+direction)/2)
-            if bracket_param > self.bracket_strategy[self.iparam]['nonidentifiable_cutoffs'][ib]:
-                self.vprint(2, "{} edge of confidence interval is non-identifiable".format('left' if direction == -1 else 'right'))
-                p = self.fit.bounds[self.iparam][ib]
+            # Check bounds
+            past_bound = direction*(p - param.bounds[idir]) >= 0
+            if past_bound:
+                p = param.bounds[idir]
+
+            # Calculate profile likelihood
+            pL = self.profile_likelihood(p)
+
+            # Where next?
+            if pL <= pL_thres:
+                self.vprint(3, f"bracketing: {pL:.3f} <= {pL_thres:.3f} @ {p}")
+                break
+            elif past_bound:
+                self.vprint(3, "bracket reached {['lower', 'upper'][idir]} bound: {pL:.3f} > {pL_thres:.3f} @ {p}")
                 pL = np.inf
                 break
-
-            # Update position
-            if self.bracket_strategy[self.iparam]['multiplicative']:
-                bracket_param *= step
-                p = x0 * bracket_param**direction
             else:
-                bracket_param += step
-                p = x0 + direction*bracket_param
+                self.vprint(3, f"bracketing: {pL:.3f} > {pL_thres:.3f} @ {p}")
+                continue # for style only
 
-            # Enforce bounds
-            if direction*(p - self.fit.bounds[self.iparam][ib]) >= 0:
-                hit_bound = True
-                p = self.fit.bounds[self.iparam][ib]
+        else: # we moved too far from the point estimate; non-identifiable
+            self.vprint(2, f"{['left', 'right'][idir]} edge of confidence interval is non-identifiable (or <fit>.parameters[{self.cur_param}].max_linearization_moves is too tight)")
+            p = param.bounds[idir]
+            pL = np.inf
 
-            # Update profile likelihood
-            pL = self.profile_likelihood(p)
-        else:
-            if pL <= pL_thres:
-                self.vprint(3, "bracketing: {:.3f} < {:.3f} @ {}".format(pL, pL_thres, p))
-            else:
-                assert hit_bound # Sanity check
-                self.vprint(3, "bracket reached {} bound: {:.3f} > {:.3f} @ {}".format("lower" if direction < 0 else "upper", pL, pL_thres, p))
-                pL = np.inf
-            
         return p, pL
         
     def initial_bracket_points(self):
@@ -627,9 +538,8 @@ class Profiler():
         --------
         find_bracket_point, find_MCI
         """
-        self.expand_bracket_strategy()
-        a, a_pL = self.iterate_bracket_point(self.point_estimate['params'][self.iparam], self.point_estimate['logL'], direction=-1)
-        b, b_pL = self.iterate_bracket_point(self.point_estimate['params'][self.iparam], self.point_estimate['logL'], direction= 1)
+        a, a_pL = self.find_bracket_point(direction=-1)
+        b, b_pL = self.find_bracket_point(direction= 1)
         return (a, a_pL), (b, b_pL)
     
     def solve_bisection(self, bracket, bracket_pL):
@@ -638,8 +548,8 @@ class Profiler():
 
         Parameters
         ----------
-        bracket : [float, float]
-            the bracket of parameter values containing the root
+        lin_bracket : [float, float]
+            the bracket of parameter values containing the root.
         bracket_pL : [float, float]
             the posterior values associated with the bracket points
 
@@ -653,43 +563,52 @@ class Profiler():
         -----
         All points evaluated along the way are stored in ``self.ress``
         """
-        bracket_width = np.diff(bracket)[0]
-        if bracket_width < 1e-10:
-            # Sometimes the (profile) likelihood is discontinuous right at the
-            # (prospective) CI bound, in which case we will not converge to the
-            # LR_interval. For the CI this is no problem, it just extends to
-            # the location of the jump. To be conservative, we include the jump
-            # in the CI.
+        # Input processing
+        L = self.fit.parameters[self.cur_param].linearization
+        pe = self.point_estimate['params'][self.cur_param]
+
+        # Sometimes the (profile) likelihood is discontinuous right at the
+        # (prospective) CI bound, in which case we will not converge to the
+        # LR_interval. For the CI this is no problem, it just extends to the
+        # location of the jump. To be conservative, we include the jump in the
+        # CI.
+        if L.distance(pe, *bracket) < 1e-10:
             self.vprint(3, "bracket collapsed; likelihoods are ({:.3f}, {:.3f})".format(*bracket_pL))
             return bracket[np.argmin(bracket_pL)]
 
-        c = np.mean(bracket)
+        # Get bisector
+        c = L.mean(pe, bracket)
         c_pL = self.profile_likelihood(c)
         
-        self.vprint(3, f"current bracket: {c} +- {bracket_width/2:.5g} @ {bracket_pL[0]:.3f}, {c_pL:.3f}, {bracket_pL[1]:.3f}")
+        self.vprint(3, f"current bracket: {bracket[0]} < {c} < {bracket[1]} @ {bracket_pL[0]:.3f}, {c_pL:.3f}, {bracket_pL[1]:.3f}")
 
+        # Evaluate function values (likelihood - target)
         a_fun, b_fun = self.point_estimate['logL'] - bracket_pL - self.LR_target
         c_fun = self.point_estimate['logL'] - c_pL - self.LR_target
-        i_update = 0 if a_fun*c_fun > 0 else 1
-        assert [a_fun, b_fun][1-i_update]*c_fun <= 0
+
+        # Figure out which side of the bracket to move inwards
+        i_move = 0 if a_fun*c_fun > 0 else 1
+        assert [a_fun, b_fun][1-i_move]*c_fun <= 0
         
-        bracket[i_update] = c
-        bracket_pL[i_update] = c_pL
+        # Update the bracket
+        bracket[i_move] = c
+        bracket_pL[i_move] = c_pL
         
+        # Check whether we converged or should recurse
         if np.all([self.LR_interval[0] < LR < self.LR_interval[1] for LR in self.point_estimate['logL'] - bracket_pL]):
-            return np.mean(bracket)
+            return L.mean(pe, bracket)
         else:
             return self.solve_bisection(bracket, bracket_pL)
     
     @restart_if_better_pe_found
-    def find_single_MCI(self, iparam):
+    def find_single_MCI(self, param_name):
         """
         Run the whole profiling process for one parameter
 
         Parameters
         ----------
-        iparam : int
-            index of the parameter to sweep
+        param_name : str
+            name of the parameter to sweep
 
         Returns
         -------
@@ -703,14 +622,14 @@ class Profiler():
         --------
         find_MCI
         """
-        self.iparam = iparam
+        self.cur_param = param_name
         if self.point_estimate is None:
-            raise RuntimeError("Need to have a point estimate before calculating confidence intervals") # pragma: no cover
+            raise RuntimeError("Need to have a point estimate before calculating confidence intervals")
             
-        (a, a_pL), (b, b_pL) = self.initial_bracket_points()
-        m = self.point_estimate['params'][self.iparam]
+        m = self.point_estimate['params'][self.cur_param]
         m_pL = self.point_estimate['logL']
 
+        (a, a_pL), (b, b_pL) = self.initial_bracket_points()
         self.vprint(3, "Found initial bracket boundaries, now solving by bisection")
         
         roots = np.array([np.nan, np.nan])
@@ -719,16 +638,22 @@ class Profiler():
         if b_pL == np.inf:
             roots[1] = b
         
-        for i, (bracket, bracket_pL) in enumerate(zip([(a, m), (m, b)], [(a_pL, m_pL), (m_pL, b_pL)])):
-            if np.isnan(roots[i]):
-                roots[i] = self.solve_bisection(np.asarray(bracket), np.asarray(bracket_pL))
+        for i, (bracket, bracket_pL) in enumerate(zip([(a, m), (m, b)],
+                                                      [(a_pL, m_pL), (m_pL, b_pL)])):
+            if bracket_pL[i] == np.inf: # [i] is the outer one in both cases
+                roots[i] = bracket[i]
+            else:
+                roots[i] = self.solve_bisection(np.asarray(bracket),
+                                                np.asarray(bracket_pL))
+
             if i == 0:
-                self.vprint(2, f"Found left edge @ {roots[i]}, now searching right one")
+                self.vprint(2, f"Identified left edge @ {roots[0]}, now looking for right one")
         
-        self.vprint(2, f"found CI = {roots} (point estimate = {m}, iparam = {self.iparam})\n")
+        self.vprint(2, f"found CI = {roots} (point estimate = {m}, current parameter is {self.cur_param})\n")
         return m, roots
     
-    def find_MCI(self, iparam='all',
+    def find_MCI(self,
+                 parameters=all, # clear, but a bit abusive
                  show_progress=False,
                 ):
         """
@@ -736,21 +661,22 @@ class Profiler():
 
         Parameters
         ----------
-        iparam : 'all' or np.ndarray, dtype=int
-            the parameters to sweep. If 'all', will sweep all independent
-            parameters, i.e. those not fixed via the `Fit.fix_values`
-            mechanism.
+        parameters : str or list of str, optional
+            the names of the parameters to sweep. Defaults to "all independent
+            parameters from the fit"
 
         Returns
         -------
-        mcis : (n, 3) np.ndarray, dtype=float
-            for each parameter, in order: point estimate, lower, and upper
-            bound.
+        dict
+            for each parameter a tuple ``(m, ci)``, where ``m`` is the point
+            estimate and ``ci`` a numpy array with lower and upper bound; c.f.
+            `find_single_MCI`
 
         See also
         --------
-        run_fit
+        run_fit, find_single_MCI
         """
+        # Input processing
         if show_progress and self.bar is None:
             self.bar = tqdm() # pragma: no cover
             
@@ -762,32 +688,29 @@ class Profiler():
                                                                                  self.point_estimate['logL'],
                                                                                 ))
         
-        n_params = len(self.point_estimate['params'])
-        mcis = np.empty((n_params, 3), dtype=float)
-        mcis[:] = np.nan
-        
-        # check, which parameters to run
-        if iparam == 'all':
-            fixed = [i for i, _ in self.fit.fix_values]
-            iparams = np.array([i for i in range(n_params) if i not in fixed])
-        else:
-            iparams = np.asarray(iparam)
-            if len(iparams.shape) == 0:
-                iparams = np.array([iparams])
+        if parameters is all:
+            parameters = self.fit.independent_parameters()
+        elif isinstance(parameters, str):
+            parameters = [parameters]
                
         # keep track of which point estimate was used for which parameters
-        used_point_estimates = n_params*[None]
+        used_point_estimate = {}
+
+        mcis = {}
         while True: # limited by max_restarts_per_parameter
-            for iparam in iparams:
-                # Only run if we did not already sweep from here
-                if used_point_estimates[iparam] is not self.point_estimate:
+            for cur_param in parameters:
+                try:
+                    ran_from_here = used_point_estimate[cur_param] is self.point_estimate
+                except KeyError:
+                    ran_from_here = False
+
+                if not ran_from_here:
                     self.run_count = 0
-                    self.vprint(2, f"starting iparam = {iparam}")
-                    m, ci = self.find_single_MCI(iparam)
-                    mcis[iparam, :] = m, *ci
-                    used_point_estimates[iparam] = self.point_estimate
+                    self.vprint(2, f"starting parameter {cur_param}")
+                    mcis[cur_param] = self.find_single_MCI(cur_param)
+                    used_point_estimate[cur_param] = self.point_estimate
                 
-            if ( all([used_point_estimates[iparam] is self.point_estimate for iparam in iparams])
+            if ( all(used_pe is self.point_estimate for _, used_pe in used_point_estimate.items())
                  or not self.restart_on_better_point_estimate ):
                 break
         
@@ -796,8 +719,4 @@ class Profiler():
             self.bar = None
             
         self.vprint(2, "Done\n")
-        
-        if len(iparams) == 1:
-            return mcis[iparams[0]]
-        else:
-            return mcis
+        return mcis
